# -*- coding: utf-8 -*-
""" Various helpers used by the serialization library. """
from __future__ import absolute_import, print_function, unicode_literals

# stdlib imports
import types

# 3rd party imports
from six import string_types


def iterable(collection):
    """Checks if the variable is not a basestring instance and can be
    enumerated.
    """
    # strings can be iterated - that's not what we want
    if isinstance(collection, string_types):
        return False

    # avoid opening a generator
    if isinstance(collection, types.GeneratorType):
        return True

    try:
        iter(collection)
    except TypeError:
        return False

    return True


def is_file(obj):
    """ Check if the given object is file-like. """
    return hasattr(obj, 'flush') and hasattr(obj, 'readline')



# -*- coding: utf-8 -*-
"""
Built-in serializers for most basic uses cases.

Those are provided as they would have to be reimplemented time and time again
by each project. Having them as part of the library will also allow some
performance optimisation in the `Serializer` itself.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
from datetime import date as Date, datetime as Datetime
from decimal import Decimal
from itertools import chain
from logging import getLogger

# 3rd party imports
from six import integer_types, string_types, binary_type

# local imports
from .core import serialize
from .util import is_file


L = getLogger(__name__)
PRIMITIVES = tuple(chain(
    integer_types,
    string_types,
    (float, binary_type, Date, Datetime, Decimal)
))


def serialize_dict(dct, spec, ctx):
    """ Serialize dictionary. """
    ret = {}

    if spec is True or spec.empty():
        return {}

    for name, value in dct.items():
        if name in spec:
            try:
                ret[name] = serialize.raw(value, spec[name], ctx)
            except ValueError:
                pass
    return ret


def serialize_primitive(obj, spec, ctx):
    """ Serialize a primitive value. """
    return ctx.dumpval('', obj)


def serialize_iterable(obj, spec, ctx):
    """ Serialize any iterable except a string.

    Since strings are a very special case of iterables, they are handled
    differently (otherwise they would be serialized as an array of chars).
    """
    ret = []
    for item in obj:
        ret.append(serialize.raw(item, spec, ctx))
    return ret


def serialize_file_handle(obj, spec, ctx):
    """ Serializer for file handles. """
    return '(file handle)'


def serialize_serializable(obj, spec, ctx):
    """ Serialize any class that defines a ``serialize`` method. """
    return obj.serialize(spec, ctx)


class ThirdPartySerializer(object):
    """ A serializer for integrating with 3rd party solutions.

    This will use a given method on the object to return the dict and then
    use `serialize_dict` to filter the results according to the specification.

    This will be slower than implementing a direct `serafin_serialize()` method
    but will work with things like AppEngine ndb models ``to_dict()`` method out
    of the box.
    """
    def __init__(self, method_name):
        self.method_name = method_name

    def __call__(self, obj, spec, ctx):
        method = getattr(obj, self.method_name)
        data = method()
        return serialize_dict(data, spec, ctx)


def serialize_object(obj, spec, ctx):
    """ Serialize any object.

    This should have the lowest priority as it will work for almost anything
    but it might be a bit slow. It's generally much better to use other,
    dedicated, serializers and leave this as a last resort. Having this makes
    the system very flexible but at the same time it might slow down the
    serialization time significantly.
    """
    if spec is True or spec.empty():
        return {}

    filters = [
        lambda n, v: not (n.startswith('__') or n.endswith('__')),
        lambda n, v: not callable(v),
        lambda n, v: not is_file(v),
    ]

    def is_val(attr_name, attr_value):
        """
        Check if the given attribute is a value that should be serialized.
        """
        try:
            return all(flt(attr_name, attr_value) for flt in filters)
        except:
            return False

    ret = {}
    for name in dir(obj):
        if not isinstance(name, string_types):
            pass
        if name in spec:
            try:
                value = getattr(obj, name)
            except Exception as ex:
                value = "({}: {})".format(ex.__class__.__name__, str(ex))
                if ctx.reraise:
                    raise

            if is_val(name, value):
                ret[name] = serialize.raw(
                    value, spec[name], ctx
                )
    return ret



# -*- coding: utf-8 -*-
"""
Serialization context implementation.
"""
from __future__ import absolute_import, unicode_literals
import json
from datetime import date, datetime
from decimal import Decimal

from six import string_types, iteritems


def json_dump_default(value):
    """ Default JSON serializer. """
    if isinstance(value, Decimal):
        return float(value)

    elif isinstance(value, (date, datetime)):
        return value.isoformat()

    elif hasattr(value, 'serialize') and callable(value.serialize):
        return value.serialize()

    return value


class Context(dict):
    """ Serialization context.

    It allows easy reading/writing of values into it and can be easily
    serialized to/from JSON.
    """
    @classmethod
    def create(cls, data=None):
        """ Create new context.

        :param dict|string data:
            Initial context data. This will be shallow copied into the context.
            This can be either a string with the JSON object or anything that
            is accepted by the built-in dict() constructor.
        :return:
            Newly created context object.
        """
        if data is None:
            return cls()
        elif isinstance(data, string_types):
            return cls.loads(data)
        else:
            ret = cls()
            for name, value in iteritems(data):
                if isinstance(value, dict):
                    value = Context.create(value)
                ret[name] = value
            return ret

    @classmethod
    def loads(cls, string):
        """ Load context from JSON string. """
        return json.loads(string, object_pairs_hook=cls)

    def __repr__(self):
        """ Return pretty representation for the context. """
        return '<Context {0}>'.format(str(self))

    def __str__(self):
        """ Use JSON as string representation """
        return json.dumps(self, indent=2, default=json_dump_default)

    def __unicode__(self):
        return str(self).encode('utf-8')

    def __setattr__(self, name, value):
        """ This allows setting context values as if they were attributes. """
        self[name] = value

    def __getattr__(self, name):
        """ This allows reading context values as if they were attributes.

        This makes it a bit easier to use the Context objects (they work a bit
        like JavaScript objects).
        """
        try:
            return self[name]
        except KeyError:
            raise AttributeError("No attribute {}".format(name))



# -*- coding: utf-8 -*-
"""

**serafin** is a serialization system that allows flexible serialization
of any type of object according to a provided field spec. The field spec tells
the serialize which attribute/fields/members of the given object should be
serialized. This allows for a very flexible serialization system, especially in
the context of API endpoints where we can write one endpoint and allow client
to pass the field spec describing how he wants the output to be formatted.


.. autoclass:: serafin.core.Priority
    :members:


.. autoclass:: serafin.core.Serializer
    :members:


.. autofunction:: serafin.core.serialize

"""
from __future__ import absolute_import, unicode_literals
from .context import Context
from .core import serialize
from .fieldspec import Fieldspec
from .serializers import *     # pylint: disable=wildcard-import


__all__ = [
    'Context',
    'Fieldspec',
    'serialize',
]



# -*- coding: utf-8 -*-
""" Serialization field spec implementation.

This is what makes the serialization so flexible as it allows to cherry-pick
what is actually serialized. This module just implements the serialization field
specification itself and the rest of the system uses it to define/decide what
is being included in the resulting data.

This module **has** to be really well tested as everything else depends on it
and the implementation itself is a bit tricky.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import sys
import re
from copy import deepcopy
from collections import OrderedDict
if sys.version_info >= (3, 0):
    from urllib.parse import parse_qsl
else:
    from urlparse import parse_qsl

# 3rd party imports
from six import string_types, iteritems     # noqa


kFieldPtrn = re.compile(r'{name}(\({members}\))?'.format(
    name=r'(?P<name>-?[\w\d_*]+)',
    members=r'(?P<members>[\w\d_\-,*)(]+)'
))


class Fieldspec(object):
    """
    ``Fieldspec`` represents field specification for REST queries. It describes
    the format in which the query result will be returned. What model fields
    should be present and which of them should be expanded. Expansion works
    for foreign keys, where instead of object key, we can insert the object
    itself (with all, or selected members).

    For usage examples see ``restapi.serialize.serialize()``.

    **Simple tests**

    >>> fs = Fieldspec('field1,field2(sub1,sub2),field3(*)')
    >>> fs
    <Fieldspec: field1,field2,field3>

    Easily check if the field is matched by the spec:
    >>> 'field1' in fs
    True
    >>> fs['field1']
    True
    >>> 'field7' in fs
    False

    Supports nested specifications:
    >>> fs['field2']
    <Fieldspec: sub1,sub2>
    >>> 'sub1' in fs['field2']
    True
    >>> fs['field2']['sub2']
    True

    And wildcards
    >>> 'asdf' in fs['field2']
    False
    >>> 'asdf' in fs['field3']
    True
    >>> fs['field3']['asdf']
    True

    Exclude
    >>> fs = Fieldspec('*,-field1')
    >>> 'asdf' in fs
    True
    >>> 'field1' in fs
    False

    **Implementation**

    Fieldspec has 2 main variables that hold the spec. The ``fields`` and
    ``exclude``. Exclude is much simpler, because it's just a list of strings
    with field names that should be excluded at this level.

    The ``fields`` member is a list of tuples ``(name, members)`` where name
    is the fields name and members is the optional member spec **members** can
    be one of:
        - None
        - ``True`` if the subfield spec is just the field name, ie ``field``.
        - ``Fieldspec`` if the subfield has a field spec defined, ie
         ``field(sub1,sub2)``.

    If the field spec contains ``*`` it means at that level all fields are
    included by default. You can still exclude them using ``-`` but by default
    all fields within an object are included.

    The special field spec `**` can be used and it means the same thing as ``*``
    but is applied recursively to all objects. **Be very careful with using it
    as it will try to serialize EVERYTHING and in many cases you will run into
    problems.

    TODO:
        - Ability to specify max depth.
        - Merging two specs together (union).
        - Restricting one spec by another (intersection).
    """
    def __init__(self, spec='*'):
        self.fields = OrderedDict()
        self.exclude = OrderedDict()
        self.spec = spec
        self.all = False

        if isinstance(spec, Fieldspec):
            # Clone
            self.fields = deepcopy(spec.fields)
            self.exclude = deepcopy(spec.exclude)
            self.all = spec.all
            self.spec = spec.spec

        elif isinstance(spec, string_types):
            self._parse(spec)

        elif spec not in (None, True):
            raise ValueError('Invalid field specification')

    def empty(self):
        """ Return **True** if the current spec is empty.

        :return bool:
            **True** if nothing represented by this field spec should be
            included as part of the serialization result.
        """
        return len(self.fields) == 0 and not self.all

    def __contains__(self, name):
        if name in self.exclude:
            return False

        if self.all:
            return True

        try:
            return self.fields[name]
        except KeyError:
            return None

    def __getitem__(self, name):
        if name in self.exclude:
            return None

        if self.spec == '**':
            return Fieldspec(self)

        try:
            return self.fields[name]
        except KeyError:
            return self.all or None

    def __repr__(self):
        include = ','.join(name for name in self.fields.keys())
        exclude = ''
        if self.exclude:
            exclude = ','.join('-' + n for n, m in self.exclude)

        return "<Fieldspec: {}{}>".format(include, exclude)

    def _parse(self, string):
        """ Parse the spec string.
        Args:
            string (str):   A string representation of the field spec.

        The string is split on the high level by commas and then processed.
        The parser will take into account the parenthesis and skip them when
        looking for comas. The members are parsed recursively.
        """
        fields  = self._splitfields(string)
        for field in fields:
            m = kFieldPtrn.match(field)
            if not m:
                raise ValueError("Invalid field specification")

            name = m.group('name')
            if name[0] == '-':
                name    = name[1:]
                group   = self.exclude
            else:
                group   = self.fields

            membstr = m.group('members')
            members = True if membstr is None else Fieldspec(membstr)
            if name in ('*', '**'):
                self.all = True
            else:
                group[name] = members

    def _splitfields(self, string):
        """
        Split field list into separate fields. Takes into account nesting
        specification (inside parenthesis) by ignoring commas inside it.

        >>> fs = Fieldspec()
        >>> fs._splitfields('one,two,three')
        ['one', 'two', 'three']

        >>> fs._splitfields('one(mem1,mem2),two,three')
        ['one(mem1,mem2)', 'two', 'three']

        """
        fields = []
        append = fields.append
        start = 0
        numparen = 0

        for i, ch in enumerate(string):
            if ch == ',':
                if i != start and numparen == 0:
                    append(string[start:i])
                    start = i + 1
            elif ch == '(':
                numparen += 1
            elif ch == ')':
                numparen -= 1

        if numparen != 0:
            raise ValueError("Unmatched parenthesis")

        fields.append(string[start:])
        return fields

    def merge(self, spec):
        """ Extend field spec using a different one. The result should be
        the union of both.

        .. notes:: If the field is in both specs and it's values are different
        there is a conflict. The conflict can be resolved by merging the sub
        specs.
        """
        if isinstance(spec, string_types):
            spec = Fieldspec(spec)

        if spec.all:
            self.all = True

        for name, members in iteritems(spec.fields):
            try:
                mymembers = self.fields[name]
                # There already is a field like that, we need to sort
                # out the conflict.
                if mymembers != members:
                    self.fields[name] = self._resolve_conflict(
                        mymembers, members
                    )
            except KeyError:
                # this field spec doesn't have the key yet
                self.fields[name] = members

        exclude = self.exclude
        self.exclude = OrderedDict()
        for name, members in iteritems(exclude):
            try:
                members = spec.fields[name]
                self.fields[name] = members
            except KeyError:
                self.exclude[name] = members

        return self

    def _resolve_conflict(self, mymembers, members):
        """ Resolve conflicting members by mergin them. """
        if isinstance(mymembers, Fieldspec):
            if isinstance(members, Fieldspec):
                return Fieldspec(mymembers).merge(members)
            else:  # members == True:
                return mymembers
        else:  # mymembers == True:
            return members

    def restrict(self, spec):
        """ Restrict the current field spec by another one. The result should be
        the intersection of both.
        """
        was_all = self.all
        my_fields = frozenset(self.fields.keys())
        other_fields = frozenset(spec.fields.keys())

        if spec.all:
            common = my_fields
        elif self.all:
            self.all = False
            common = other_fields
        else:
            common = my_fields & other_fields

        self.exclude.update(spec.exclude)

        newfields = []
        for name in common:
            try:
                mymembers = self.fields[name]
            except KeyError:
                if was_all:
                    newfields.append((name, spec.fields[name]))
                continue

            try:
                members = spec.fields[name]
                if mymembers != members:
                    newfields.append((name, self._resolve_restrict_conflict(
                        mymembers, members
                    )))
                    continue
            except KeyError:
                pass

            if name not in self.exclude:
                newfields.append((name, mymembers))

        self.fields = OrderedDict(newfields)
        return self

    def _resolve_restrict_conflict(self, mymembers, members):
        if isinstance(mymembers, Fieldspec):
            if isinstance(members, Fieldspec):
                return Fieldspec(mymembers).restrict(members)
            if members is True:
                return True
        raise NotImplementedError("Not implemented")

    @classmethod
    def from_query(self, qs):
        """ Create a field spec from a HTTP request query string. """
        qs  = dict(parse_qsl(qs)) if isinstance(qs, string_types) else qs
        return Fieldspec(qs.get('_fields', '*'))



# -*- coding: utf-8 -*-
"""
Contains the core serializer functionality.
"""
from __future__ import absolute_import

# stdlib imports
from collections import OrderedDict
from datetime import date as Date, datetime as Datetime
from logging import getLogger

# 3rd party imports
from six import iteritems, string_types

# local imports
from .context import Context
from .fieldspec import Fieldspec
from .util import iterable, is_file


L = getLogger(__name__)


def dump_val(name, value):
    """ The default implementation for object dump passed to Context. """
    if isinstance(value, (Date, Datetime)):
        return value.isoformat()
    return value


class Serializer(object):
    """ The serializer implementation

    This will serialize the object based on the field spec passed.

    Args:
        obj (anything):
            The serialized object. Whether the object will be serialized
            depends on if there is a serializer defined for that object.
        spec (Fieldspec or str):
            Fieldspec according to which the object will be serialized.
        dumpval (Function):
            The value dumping function. This will be used to serialize
            primitive types.

    Returns:
        An object representation made only from primitive types. This can be
        passed to a function like json.dumps to dump it to a output format of
        choice.

    **Examples**

    With the following data (same applies to django models and ``Context``):

    >>> model = {
    ...     'field1': 10,
    ...     'field2': {
    ...         'sub1': 1,
    ...         'sub2': 2,
    ...     },
    ...     'field3': 20,
    ... }

    Here are a few examples of what fields would be selected by each
    spec (second argument for ``serialize``):

    >>> from serafin import serialize
    >>> serialize(model, '*') == {
    ...     'field1': 10,
    ...     'field2': {},
    ...     'field3': 20
    ... }
    True

    Serialize only selected fields.

    >>> serialize(model, 'field1,field3') == {
    ...     'field1': 10,
    ...     'field3': 20
    ... }
    True

    Specify what fields to expand on an object.

    >>> serialize(model, 'field1,field2(sub1)') == {
    ...     'field1': 10,
    ...     'field2': {
    ...         'sub1': 1
    ...     }
    ... }
    True

    Wildcards (``*``) expand all fields for that given object, *without*
    expanding nested objects.

    >>> serialize(model, 'field1,field2(*)') == {
    ...     'field1': 10,
    ...     'field2': {
    ...         'sub1': 1,
    ...         'sub2': 2
    ...     }
    ... }
    True

    Double wildcard (``**``) will expand all fields recursively. This is the
    most heavy call.

    >>> serialize(model, '**') == {
    ...     'field1': 10,
    ...     'field2': {
    ...         'sub1': 1,
    ...         'sub2': 2
    ...     },
    ...     'field3': 20
    ... }
    True

    """

    def __init__(self):
        self.classmap = OrderedDict()

    def type(self, cls):
        """ Decorator for serializers that match strictly by type.

        Each type can have only one strict serializer associated with it.
        """
        def decorator(fn):      # pylint: disable=missing-docstring
            if cls in self.classmap:
                L.warning("Strict serializer already registered for {}".format(
                    cls.__name__
                ))
            self.classmap[cls] = fn
            return fn
        return decorator

    def __call__(self, obj, spec='*', **kwargs):
        """ Serialize the object according to the given field spec.

        This is the method that should be used by the users of the lib.
        Compared to `raw_serialize()` it will convert field spec to a
        `Fieldspec` instance if necessary and build the serialization
        context based on the extra keyword arguments passed.

        :param Any obj:
        :param Fieldspec|unicode|str spec:
        :param dict kwargs:
        :return dict.:
            Returns an object that can be directly dumped to JSON.
        """
        if isinstance(spec, string_types):
            spec = Fieldspec(spec)
        elif spec is None:
            spec = Fieldspec('*')

        ctx = Context(
            dumpval=dump_val,
            reraise=True,
        )
        ctx.update(kwargs)

        return self.raw(obj, spec, ctx)

    def raw(self, obj, spec, ctx):
        """ Raw serialize without parsing the field spec.

        This method should be used when writing new serializers for performance
        reasons. When writing a new serializer you will already have a
        serialization context passed to the serializer function and the
        field spec will also already be passed as `Fieldspec` instance. Using
        `raw_serializer` will remove the overhead of creating those at
        each call level. The `serialize` method will create those when executed
        by the user code and then pass them to the serializer function. That
        function in turn might need to use other serializers defined elsewhere
        but won't need the overhead of calling `serialize()` again.

        :param Any obj:
        :param Fieldspec spec:
        :param dict ctx:
        :return dict|list|str|int:
            Returns an object that can be directly dumped to JSON.
        """
        # Find serializer (inlined for performance reasons
        if obj is None:
            return None

        # Check if we have a direct serializer for the class
        try:
            serializer = self.classmap[obj.__class__]
        except KeyError:
            # Hacky speedups
            if isinstance(obj, dict):
                serializer = serialize_dict
            elif isinstance(obj, PRIMITIVES):
                serializer = serialize_primitive
            elif is_file(obj):
                serializer = serialize_file_handle
            elif iterable(obj):
                serializer = serialize_iterable
            elif callable(getattr(obj, 'serafin_serialize', None)):
                serializer = serialize_serializable
            else:
                # Look if we have direct serializer for a base class of obj
                for c, fn in iteritems(self.classmap):
                    if isinstance(obj, c):
                        serializer = fn
                        break
                else:
                    # Try if we can find 3rd party serializers.
                    if callable(getattr(obj, 'as_dict', None)):
                        serializer = ThirdPartySerializer('as_dict')
                    elif callable(getattr(obj, 'to_dict', None)):
                        serializer = ThirdPartySerializer('as_dict')
                    else:
                        # As a last resort try generic serialize_object
                        serializer = serialize_object

        # Do the actual serialization
        try:
            out = serializer(obj, spec, ctx)
        except Exception as ex:
            out = "({}: {})".format(ex.__class__.__name__, str(ex))
            if ctx.get('reraise', True):
                raise

        return out


serialize = Serializer()


from .serializers import (     # noqa
    PRIMITIVES,
    serialize_dict,
    serialize_file_handle,
    serialize_iterable,
    serialize_object,
    serialize_primitive,
    serialize_serializable,
    ThirdPartySerializer,
)
